import mitsuba as mi
import numpy as np
import pytest

import eradiate
from eradiate import unit_registry as ureg
from eradiate.experiments import CanopyAtmosphereExperiment
from eradiate.scenes.atmosphere import HomogeneousAtmosphere
from eradiate.scenes.biosphere import DiscreteCanopy
from eradiate.scenes.measure import MultiDistantMeasure
from eradiate.scenes.surface import CentralPatchSurface
from eradiate.test_tools.types import check_scene_element


def test_canopy_atmosphere_experiment_construct_default(mode_mono):
    """
    CanopyAtmosphereExperiment initialises with default params in all modes
    """
    exp = CanopyAtmosphereExperiment()

    # Check that the atmosphere's geometry is overridden by the experiment's
    assert exp.atmosphere.geometry is exp.geometry


def test_canopy_atmosphere_experiment_construct_measures(mode_mono):
    """
    A variety of measure specifications are acceptable
    """

    # Init with a single measure (not wrapped in a sequence)
    assert CanopyAtmosphereExperiment(measures=MultiDistantMeasure())

    # Init from a dict-based measure spec
    # -- Correctly wrapped in a sequence
    assert CanopyAtmosphereExperiment(measures=[{"type": "distant"}])

    # -- Not wrapped in a sequence
    assert CanopyAtmosphereExperiment(measures={"type": "distant"})


@pytest.mark.parametrize("padding", (0, 1))
def test_canopy_atmosphere_experiment_construct_normalize_measures(mode_mono, padding):
    # When canopy is not None, measure target matches canopy unit cell
    exp = CanopyAtmosphereExperiment(
        atmosphere=None,
        canopy=DiscreteCanopy.homogeneous(
            lai=3.0,
            leaf_radius=0.1 * ureg.m,
            l_horizontal=10.0 * ureg.m,
            l_vertical=2.0 * ureg.m,
            padding=padding,
        ),
        measures=MultiDistantMeasure(),
    )
    target = exp.measures[0].target
    canopy = exp.canopy
    assert np.isclose(target.xmin, -0.5 * canopy.size[0])
    assert np.isclose(target.xmax, 0.5 * canopy.size[0])
    assert np.isclose(target.ymin, -0.5 * canopy.size[1])
    assert np.isclose(target.ymax, 0.5 * canopy.size[1])
    assert np.isclose(target.z, canopy.size[2])

    # The measure target does not depend on the atmosphere
    exp = CanopyAtmosphereExperiment(
        geometry={"type": "plane_parallel", "width": 42.0 * ureg.km},
        atmosphere=HomogeneousAtmosphere(),
        canopy=DiscreteCanopy.homogeneous(
            lai=3.0,
            leaf_radius=0.1 * ureg.m,
            l_horizontal=10.0 * ureg.m,
            l_vertical=2.0 * ureg.m,
            padding=padding,
        ),
        measures=MultiDistantMeasure(),
    )
    target = exp.measures[0].target
    canopy = exp.canopy
    assert np.isclose(target.xmin, -0.5 * canopy.size[0])
    assert np.isclose(target.xmax, 0.5 * canopy.size[0])
    assert np.isclose(target.ymin, -0.5 * canopy.size[1])
    assert np.isclose(target.ymax, 0.5 * canopy.size[1])
    assert np.isclose(target.z, canopy.size[2])


@pytest.mark.parametrize("padding", (0, 1))
def test_canopy_atmosphere_experiment_kernel_dict(mode_mono, padding):
    # Surface width is appropriately inherited from canopy, when no atmosphere
    # is present
    exp = CanopyAtmosphereExperiment(
        atmosphere=None,
        canopy=DiscreteCanopy.homogeneous(
            lai=3.0,
            leaf_radius=0.1 * ureg.m,
            l_horizontal=10.0 * ureg.m,
            l_vertical=2.0 * ureg.m,
            padding=padding,
        ),
        measures=[
            {"type": "distant", "id": "distant_measure"},
            {"type": "radiancemeter", "origin": [1, 0, 0], "id": "radiancemeter"},
        ],
    )

    mi_wrapper = check_scene_element(
        exp.scene, mi.Scene, drop_parameters=False
    )  # Do not drop untracked parameters: we want to check the surface transform
    assert np.allclose(
        mi_wrapper.parameters["surface_shape.to_world"].transform_affine(
            mi.Point3f(1, -1, 0)
        ),
        [5 * (2 * padding + 1), -5 * (2 * padding + 1), 0],
    )

    # -- Measures get no external medium assigned
    assert all(sensor.medium() is None for sensor in mi_wrapper.obj.sensors())

    # Surface width is appropriately inherited from atmosphere
    exp = CanopyAtmosphereExperiment(
        geometry={"type": "plane_parallel", "width": 42.0 * ureg.km},
        atmosphere=HomogeneousAtmosphere(),
        canopy=DiscreteCanopy.homogeneous(
            lai=3.0,
            leaf_radius=0.1 * ureg.m,
            l_horizontal=10.0 * ureg.m,
            l_vertical=2.0 * ureg.m,
            padding=padding,
        ),
    )

    mi_wrapper = check_scene_element(
        exp.scene, mi.Scene, drop_parameters=False
    )  # Do not drop untracked parameters: we want to check the surface transform
    assert np.allclose(
        mi_wrapper.parameters["surface_shape.to_world"].matrix,
        mi.ScalarTransform4f.scale([21000, 21000, 1]).matrix,
    )


@pytest.mark.slow
def test_canopy_atmosphere_experiment_surface_adjustment(mode_mono):
    """
    Create a CanopyAtmosphereExperiment and assert the central patch surface is created
    with the correct parameters, according to the canopy and atmosphere.
    """
    exp = CanopyAtmosphereExperiment(
        geometry={"type": "plane_parallel", "width": 42.0 * ureg.km},
        atmosphere=HomogeneousAtmosphere(),
        canopy=DiscreteCanopy.homogeneous(
            lai=3.0,
            leaf_radius=0.1 * ureg.m,
            l_horizontal=10.0 * ureg.m,
            l_vertical=2.0 * ureg.m,
        ),
        surface=CentralPatchSurface(
            bsdf={"type": "lambertian"},
            patch_bsdf={"type": "rpv"},
            patch_edges=10.0 * ureg.m,
        ),
    )

    expected = (
        mi.ScalarTransform3f.scale([1400, 1400])
        @ mi.ScalarTransform3f.translate([-0.499642857, -0.499642857])
    ).matrix

    mi_wrapper = check_scene_element(
        exp.scene, mi.Scene, drop_parameters=False
    )  # Do not drop untracked parameters: we want to check the surface texture's uv transform
    result = mi_wrapper.parameters["surface_bsdf.weight.to_uv"].matrix

    np.testing.assert_allclose(expected, result)


@pytest.mark.slow
def test_canopy_atmosphere_experiment_real_life(
    mode_mono,
    atmosphere_us_standard_mono,
):
    # Construct with typical parameters
    exp = CanopyAtmosphereExperiment(
        surface={"type": "rpv"},
        atmosphere={
            "type": "heterogeneous",
            "molecular_atmosphere": atmosphere_us_standard_mono,
        },
        canopy={
            "type": "discrete_canopy",
            "construct": "homogeneous",
            "lai": 3.0,
            "leaf_radius": 0.1 * ureg.m,
            "l_horizontal": 10.0 * ureg.m,
            "l_vertical": 2.0 * ureg.m,
        },
        illumination={"type": "directional", "zenith": 45.0},
        measures=[
            {
                "type": "distant",
                "construct": "hplane",
                "zeniths": np.arange(-60, 61, 5),
                "azimuth": 0.0,
                "id": "distant",
            },
            {"type": "radiancemeter", "origin": [1, 0, 0], "id": "radiancemeter"},
        ],
    )
    mi_wrapper = check_scene_element(exp.scene, mi.Scene, ctx=exp.context_init)

    # -- Distant measures get no external medium
    assert mi_wrapper.obj.sensors()[0].medium() is None

    # -- Radiancemeter inside the atmosphere must have a medium assigned
    assert mi_wrapper.obj.sensors()[1].medium().id() == "medium_atmosphere"


@pytest.mark.slow
def test_canopy_atmosphere_experiment_run_detailed(mode_mono):
    """
    Test for correctness of the result dataset generated by CanopyAtmosphereExperiment.
    Note: This test is outdated, most of its content should be transferred to
    tests for measure post-processing pipelines.
    """
    exp = CanopyAtmosphereExperiment(
        measures=[
            {
                "id": "toa_brf",
                "type": "distant",
                "construct": "hplane",
                "zeniths": np.arange(-60, 61, 5),
                "azimuth": 0.0,
            },
        ]
    )
    results = eradiate.run(exp)

    # Post-processing creates expected variables ...
    assert set(results.data_vars) == {"brdf", "brf", "irradiance", "radiance"}

    # ... dimensions
    assert set(results["radiance"].dims) == {"saa", "sza", "w", "x_index", "y_index"}
    assert set(results["irradiance"].dims) == {"saa", "sza", "w"}

    # ... and other coordinates
    expected_coords = {"saa", "sza", "vaa", "vza", "x", "x_index", "y", "y_index", "w"}
    assert set(results["radiance"].coords) == expected_coords
    assert set(results["irradiance"].coords) == {"saa", "sza", "w"}

    # We just check that we record something as expected
    assert np.all(results["radiance"].data > 0.0)


def test_canopy_atmosphere_experiment_inconsistent_multiradiancemeter(
    mode_mono,
    atmosphere_us_standard_mono,
):
    # A MultiRadiancemeter measure must have all origins inside the atmosphere
    # or none. A mix of both will raise an error.

    # Construct with typical parameters
    exp = CanopyAtmosphereExperiment(
        surface={"type": "rpv"},
        atmosphere={
            "type": "heterogeneous",
            "molecular_atmosphere": atmosphere_us_standard_mono,
        },
        canopy={
            "type": "discrete_canopy",
            "construct": "homogeneous",
            "lai": 3.0,
            "leaf_radius": 0.1 * ureg.m,
            "l_horizontal": 10.0 * ureg.m,
            "l_vertical": 2.0 * ureg.m,
        },
        illumination={"type": "directional", "zenith": 45.0},
        measures=[
            {
                "type": "multi_radiancemeter",
                "origins": [[0, 0, 1], [0, 0, 1000000]],
                "directions": [[0, 0, -1], [0, 0, -1]],
                "id": "multi_radiancemeter",
            },
        ],
    )
    with pytest.raises(
        ValueError,
        match="Inconsistent placement of MultiRadiancemeterMeasure origins",
    ):
        exp.init()



import mitsuba as mi
import numpy as np
import pint
import pytest
import xarray as xr

from eradiate import KernelContext
from eradiate.constants import EARTH_RADIUS
from eradiate.scenes.core import Scene, traverse
from eradiate.scenes.geometry import PlaneParallelGeometry, SphericalShellGeometry
from eradiate.scenes.shapes import RectangleShape, SphereShape
from eradiate.scenes.surface import DEMSurface
from eradiate.scenes.surface._dem import (
    _generate_dem_vertices,
    _generate_face_indices,
    _mean_coordinates,
    _minmax_coordinates,
    _to_uv,
    _transform_vertices_plane_parallel,
    _transform_vertices_spherical_shell,
    _vertex_index,
    mesh_from_dem,
)
from eradiate.test_tools.types import check_scene_element
from eradiate.units import symbol
from eradiate.units import unit_registry as ureg


def make_dataarray(data: pint.Quantity, coords: dict):
    if set(coords.keys()) == {"lat", "lon"}:
        dims = ["lat", "lon"]
        coords = {
            "lat": (
                ["lat"],
                coords["lat"].magnitude,
                {"units": symbol(coords["lat"].units)},
            ),
            "lon": (
                ["lon"],
                coords["lon"].magnitude,
                {"units": symbol(coords["lon"].units)},
            ),
        }

    elif set(coords.keys()) == {"x", "y"}:
        dims = ["x", "y"]
        coords = {
            "x": (
                ["x"],
                coords["x"].magnitude,
                {"units": symbol(coords["x"].units)},
            ),
            "y": (
                ["y"],
                coords["y"].magnitude,
                {"units": symbol(coords["y"].units)},
            ),
        }

    else:
        raise ValueError

    return xr.DataArray(
        data=data.magnitude,
        dims=dims,
        coords=coords,
        attrs={"units": symbol(data.units)},
    )


@pytest.mark.parametrize(
    "kwargs, expected_limits",
    [
        (
            {
                "da": make_dataarray(
                    data=np.zeros((10, 10)) * ureg.m,
                    coords={
                        "lat": np.linspace(-0.1, 0.1, 10) * ureg.deg,
                        "lon": np.linspace(-0.1, 0.1, 10) * ureg.deg,
                    },
                ),
                "geometry": "plane_parallel",
                "planet_radius": EARTH_RADIUS,
            },
            [(-0.1, 0.1), (-0.1, 0.1)],
        ),
        (
            {
                "da": make_dataarray(
                    data=np.zeros((10, 10)) * ureg.m,
                    coords={
                        "lat": np.linspace(-0.2, 0.3, 10) * ureg.deg,
                        "lon": np.linspace(-0.4, 0.5, 10) * ureg.deg,
                    },
                ),
                "geometry": "spherical_shell",
                "planet_radius": EARTH_RADIUS,
            },
            [(-0.2, 0.3), (-0.4, 0.5)],
        ),
        (
            {
                "da": make_dataarray(
                    data=np.zeros((10, 10)) * ureg.m,
                    coords={
                        "x": np.linspace(-20, 20, 10) * ureg.km,
                        "y": np.linspace(-30, 30, 10) * ureg.km,
                    },
                ),
                "geometry": "plane_parallel",
                "planet_radius": EARTH_RADIUS,
            },
            [(-0.179664, 0.179664), (-0.269496, 0.269496)],
        ),
        (
            {
                "da": make_dataarray(
                    data=np.zeros((10, 10)) * ureg.m,
                    coords={
                        "x": np.linspace(-10, 20, 10) * ureg.km,
                        "y": np.linspace(-30, 40, 10) * ureg.km,
                    },
                ),
                "geometry": "spherical_shell",
                "planet_radius": EARTH_RADIUS,
            },
            [(-0.089832, 0.179664), (-0.269496, 0.359328)],
        ),
        (
            {
                "da": make_dataarray(
                    data=np.zeros((10, 10)) * ureg.m,
                    coords={
                        "x": np.linspace(-10, 20, 10) * ureg.km,
                        "y": np.linspace(-30, 40, 10) * ureg.km,
                    },
                ),
                "geometry": "invalidgeometry",
                "planet_radius": EARTH_RADIUS,
            },
            [(-0.089832, 0.179664), (-0.269496, 0.359328)],
        ),
    ],
)
def test_mesh_from_dem(modes_all_double, kwargs, expected_limits):
    if kwargs["geometry"] not in ["spherical_shell", "plane_parallel"]:
        with pytest.raises(ValueError):
            mesh_from_dem(**kwargs)
    else:
        if (kwargs["geometry"] == "spherical_shell") and kwargs[
            "planet_radius"
        ] is not None:
            with pytest.warns():
                mesh, lat, lon = mesh_from_dem(**kwargs)
        else:
            mesh, lat, lon = mesh_from_dem(**kwargs)

        assert len(mesh.vertices) == 100

        assert np.allclose(lat.m, expected_limits[0], atol=1e-4, rtol=1e-3)
        assert np.allclose(lon.m, expected_limits[1], atol=1e-4, rtol=1e-3)


@pytest.mark.parametrize(
    "geometry, expected_bg, planet_radius",
    [
        (PlaneParallelGeometry(), RectangleShape, EARTH_RADIUS),
        (SphericalShellGeometry(), SphereShape, None),
    ],
)
def test_dem_surface_construct(modes_all_mono, geometry, expected_bg, planet_radius):
    mesh, lat, lon = mesh_from_dem(
        da=make_dataarray(
            data=np.zeros((10, 10)) * ureg.m,
            coords={
                "lat": np.linspace(-0.2, 0.3, 10) * ureg.deg,
                "lon": np.linspace(-0.4, 0.5, 10) * ureg.deg,
            },
        ),
        geometry=geometry,
        planet_radius=planet_radius,
    )

    dem = DEMSurface.from_mesh(
        id="terrain", mesh=mesh, lat=lat, lon=lon, geometry=geometry
    )
    assert isinstance(dem.shape_background, expected_bg)


def test_dem_surface_kernel_dict(mode_mono):
    geometry = SphericalShellGeometry()
    mesh, lat, lon = mesh_from_dem(
        da=make_dataarray(
            data=np.zeros((10, 10)) * ureg.m,
            coords={
                "lat": np.linspace(-0.2, 0.3, 10) * ureg.deg,
                "lon": np.linspace(-0.4, 0.5, 10) * ureg.deg,
            },
        ),
        geometry=geometry,
    )

    dem = DEMSurface.from_mesh(
        id="terrain", mesh=mesh, lat=lat, lon=lon, geometry=geometry
    )

    check_scene_element(dem)

    # When enclosed in a Scene, the surface can be traversed
    scene = Scene(objects={"surface": dem})
    template, params = traverse(scene)
    kernel_dict = template.render(KernelContext())
    assert isinstance(mi.load_dict(kernel_dict), mi.Scene)


def test_generate_dem_vertices(mode_mono):
    latitude_points = np.array([1, 2])
    longitude_points = np.array([3, 4])
    elevation = np.array([[1, 2], [3, 4]])

    vertices = _generate_dem_vertices(latitude_points, longitude_points, elevation)

    expected_vertices = [[1, 3, 1], [1, 4, 2], [2, 3, 3], [2, 4, 4]]

    for exp in expected_vertices:
        assert any([np.all(exp == v) for v in vertices])


def test_generate_face_indices(mode_mono):
    face_indices = _generate_face_indices(len_x=2, len_y=2)
    expected_indices = [[3, 0, 1], [2, 0, 3]]

    for exp in expected_indices:
        assert any([np.all(exp == f) for f in face_indices])


def test_mean_coordinates(mode_mono):
    vertices = np.array([[1, 2, 3], [4, 5, 6]])

    theta_mean, phi_mean = _mean_coordinates(vertices)

    assert theta_mean == 2.5
    assert phi_mean == 3.5


def test_minmax_coordinates(mode_mono):
    vertices = np.array([[-1, -2, 3], [0, 7, 4], [6, 2, 2]])

    (theta_min, theta_max), (phi_min, phi_max) = _minmax_coordinates(vertices)

    assert theta_min == -1
    assert theta_max == 6
    assert phi_min == -2
    assert phi_max == 7


def test_vertex_index(mode_mono):
    assert _vertex_index(x=3, y=4, len_y=34) == 3 * 34 + 4


def test_transform_vertices_plane_parallel(mode_mono):
    vertices = np.array([[-1, -1, 2], [-1, 1, 2], [1, -1, 2], [1, 1, 2]])

    vertices_new = _transform_vertices_plane_parallel(vertices, 2, 10)

    expected_vertices = [
        [-0.0349, -0.0349, 12],
        [-0.0349, 0.0349, 12],
        [0.0349, -0.0349, 12],
        [0.0349, 0.0349, 12],
    ]

    for exp in expected_vertices:
        assert any([np.allclose(exp, v, atol=1e-4) for v in vertices_new])


def test_transform_vertices_spherical_shell(mode_mono):
    vertices = np.array([[-90, 0, 2], [90, 0, 2], [0, 0, 2], [0, 180, 2]])

    vertices_new = _transform_vertices_spherical_shell(vertices, 200)

    expected_vertices = [[0, 0, 202], [0, 0, -202], [202, 0, 0], [-202, 0, 0]]

    for exp in expected_vertices:
        assert any([np.allclose(exp, v, atol=1e-5) for v in vertices_new])


def test_to_uv(mode_mono):
    lat_min = -10
    lat_max = 10
    lon_min = 10
    lon_max = 30

    trafo = _to_uv(lat_min, lat_max, lon_min, lon_max)

    expected_trafo = mi.ScalarTransform4f.scale(
        (6, 3, 1)
    ) @ mi.ScalarTransform4f.translate(
        (-0.55555555556 + (0.5 / 6.0), -0.5 + (0.5 / 3.0), 0)
    )

    assert np.allclose(trafo.matrix, expected_trafo.matrix, rtol=1e-3, atol=1e-4)
